@@ -26,21 +26,23 @@
 
 #include <config.h>
 #include <ctype.h>
+#include <string.h>
 #include <glib.h>
 
 #include "gtkhtml.h"
 #include "gtkhtml-properties.h"
 
-#include "htmlobject.h"
 #include "htmlclueflow.h"
+#include "htmlcolorset.h"
 #include "htmlcursor.h"
-#include "htmllinktext.h"
+#include "htmlobject.h"
 #include "htmltable.h"
 #include "htmltext.h"
 #include "htmltextslave.h"
 #include "htmlimage.h"
 #include "htmlinterval.h"
 #include "htmlselection.h"
+#include "htmlsettings.h"
 #include "htmlundo.h"
 
 #include "htmlengine-edit.h"
@@ -89,7 +91,7 @@ html_engine_set_mark (HTMLEngine *e)
 {
 	g_return_if_fail (e != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (e));
-	g_return_if_fail (e->editable);
+	g_return_if_fail (e->editable || e->caret_mode);
 
 	if (e->mark != NULL)
 		html_engine_unselect_all (e);
@@ -160,6 +162,35 @@ html_engine_selection_pop (HTMLEngine *e)
 	html_engine_edit_selection_updater_update_now (e->selection_updater);
 }
 
+gboolean
+html_engine_selection_stack_top (HTMLEngine *e, gint *cpos, gint *mpos)
+{
+	if (e->selection_stack && GPOINTER_TO_INT (e->selection_stack->data) && e->selection_stack->next && e->selection_stack->next->next) {
+		if (cpos)
+			*cpos = GPOINTER_TO_INT (e->selection_stack->next->data);
+		if (mpos)
+			*mpos = GPOINTER_TO_INT (e->selection_stack->next->next->data);
+
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+gboolean
+html_engine_selection_stack_top_modify (HTMLEngine *e, gint delta)
+{
+	if (e->selection_stack && GPOINTER_TO_INT (e->selection_stack->data) && e->selection_stack->next && e->selection_stack->next->next) {
+		e->selection_stack->next->data = GINT_TO_POINTER (GPOINTER_TO_INT (e->selection_stack->next->data) + delta);
+		e->selection_stack->next->next->data = GINT_TO_POINTER (GPOINTER_TO_INT (e->selection_stack->next->next->data) + delta);
+
+		return TRUE;
+	}
+
+	return FALSE;
+
+}
+
 static void
 spell_check_object (HTMLObject *o, HTMLEngine *e, gpointer data)
 {
@@ -171,20 +202,29 @@ void
 html_engine_spell_check_range (HTMLEngine *e, HTMLCursor *begin, HTMLCursor *end)
 {
 	HTMLInterval *i;
+	gboolean cited;
 
 	e->need_spell_check = FALSE;
 
-	if (!e->widget->editor_api || !GTK_HTML_CLASS (GTK_OBJECT (e->widget)->klass)->properties->live_spell_check
-	    || !begin->object->parent)
+	if (!e->widget->editor_api || !gtk_html_get_inline_spelling (e->widget) || !begin->object->parent)
 		return;
 
 	begin = html_cursor_dup (begin);
 	end   = html_cursor_dup (end);
 
-	while (html_is_in_word (html_cursor_get_prev_char (begin)))
-		if (html_cursor_backward (begin, e));
-	while (html_is_in_word (html_cursor_get_current_char (end)))
-		if (html_cursor_forward (end, e));
+	cited = FALSE;
+	while (html_selection_spell_word (html_cursor_get_prev_char (begin), &cited) || cited) {
+		if (html_cursor_backward (begin, e))
+			;
+		cited = FALSE;
+	}
+
+	cited = FALSE;
+	while (html_selection_spell_word (html_cursor_get_current_char (end), &cited) || cited) {
+		if (html_cursor_forward (end, e))
+			;
+		cited = FALSE;
+	}
 
 	i = html_interval_new_from_cursor (begin, end);
 	if (begin->object->parent != end->object->parent)
@@ -213,6 +253,30 @@ html_engine_select_word_editable (HTMLEngine *e)
 		html_cursor_forward (e->cursor, e);
 }
 
+void
+html_engine_select_spell_word_editable (HTMLEngine *e)
+{
+	gboolean cited, cited2;
+
+	cited = cited2 = FALSE;
+	while (html_selection_spell_word (html_cursor_get_prev_char (e->cursor), &cited) || cited) {
+		html_cursor_backward (e->cursor, e);
+		cited2 = cited;
+		cited = FALSE;
+	}
+
+	if (cited2) {
+		html_cursor_forward (e->cursor, e);
+		cited = TRUE;
+	}
+
+	html_engine_set_mark (e);
+	while (html_selection_spell_word (html_cursor_get_current_char (e->cursor), &cited2) || (!cited && cited2)) {
+		html_cursor_forward (e->cursor, e);
+		cited2 = FALSE;
+	}
+}
+
 void
 html_engine_select_line_editable (HTMLEngine *e)
 {
@@ -300,10 +364,41 @@ html_engine_clipboard_clear (HTMLEngine *e)
 HTMLObject *
 html_engine_new_text (HTMLEngine *e, const gchar *text, gint len)
 {
-	return e->insertion_url
-		? html_link_text_new_with_len (text, len, e->insertion_font_style, e->insertion_color,
-					       e->insertion_url, e->insertion_target)
-		: html_text_new_with_len      (text, len, e->insertion_font_style, e->insertion_color);
+	HTMLObject *to;
+
+	to = html_text_new_with_len (text, len, e->insertion_font_style, e->insertion_color);
+	if (e->insertion_font_style != GTK_HTML_FONT_STYLE_DEFAULT)
+		html_text_set_style_in_range (HTML_TEXT (to), e->insertion_font_style, e, 0, HTML_TEXT (to)->text_bytes);
+	if (e->insertion_color &&
+	    e->insertion_color != html_colorset_get_color (e->settings->color_set, HTMLTextColor))
+		html_text_set_color_in_range (HTML_TEXT (to), e->insertion_color, 0, HTML_TEXT (to)->text_bytes);
+	if (e->insertion_url)
+		html_text_append_link (HTML_TEXT (to), e->insertion_url, e->insertion_target, 0, HTML_TEXT (to)->text_len);
+
+	return to;
+}
+
+HTMLObject *
+html_engine_new_link (HTMLEngine *e, const gchar *text, gint len, gchar *url)
+{
+	HTMLObject *link;
+	gchar *real_url, *real_target;
+
+	real_target = strchr (text, '#');
+	if (real_target) {
+		real_url = g_strndup (url, real_target - url);
+		real_target ++;
+	} else
+		real_url = url;
+		
+	link = html_text_new_with_len (text, len, e->insertion_font_style,
+				       html_colorset_get_color (e->settings->color_set, HTMLLinkColor));
+	html_text_append_link (HTML_TEXT (link), real_url, real_target, 0, HTML_TEXT (link)->text_len);
+
+	if (real_target)
+		g_free (real_url);
+
+	return link;
 }
 
 HTMLObject *
@@ -331,7 +426,7 @@ html_engine_get_indent (HTMLEngine *e)
 
 	return e->cursor->object && e->cursor->object->parent
 		&& HTML_OBJECT_TYPE (e->cursor->object->parent) == HTML_TYPE_CLUEFLOW
-		? HTML_CLUEFLOW (e->cursor->object->parent)->level : 0;
+		? html_clueflow_get_indentation (HTML_CLUEFLOW (e->cursor->object->parent)) : 0;
 }
 
 #define LINE_LEN 71
@@ -387,14 +482,17 @@ try_break_this_line (HTMLEngine *e, guint line_offset, guint last_space)
 }
 
 static void
-go_to_begin_of_pre_para (HTMLEngine *e)
+go_to_begin_of_para (HTMLEngine *e)
 {
 	HTMLObject *prev;
 
 	do {
+		gint offset;
 		html_cursor_beginning_of_paragraph (e->cursor, e);
-		prev = html_object_prev_leaf (e->cursor->object);
-		if (prev && html_object_get_length (prev))
+		offset = 0;
+		prev = html_object_prev_cursor (e->cursor->object, &offset);
+		if (prev && !html_object_is_container (prev) && html_object_get_length (prev)
+		    && html_clueflow_style_equals (HTML_CLUEFLOW (e->cursor->object->parent), HTML_CLUEFLOW (prev->parent)))
 			html_cursor_backward (e->cursor, e);
 		else
 			break;
@@ -402,31 +500,35 @@ go_to_begin_of_pre_para (HTMLEngine *e)
 }
 
 void
-html_engine_indent_pre_paragraph (HTMLEngine *e)
+html_engine_indent_paragraph (HTMLEngine *e)
 {
 	guint position;
 	guint line_offset;
 	guint last_space;
 
 	g_assert (e->cursor->object);
-	if (HTML_OBJECT_TYPE (e->cursor->object->parent) != HTML_TYPE_CLUEFLOW
-	    || html_clueflow_get_style (HTML_CLUEFLOW (e->cursor->object->parent)) != HTML_CLUEFLOW_STYLE_PRE)
+	if (!HTML_IS_CLUEFLOW (e->cursor->object->parent))
 		return;
 
 	html_engine_disable_selection (e);
 	position = e->cursor->position;
 
-	html_undo_level_begin (e->undo, "Indent PRE paragraph", "Reverse paragraph indentation");
+	html_undo_level_begin (e->undo, "Indent paragraph", "Reverse paragraph indentation");
 	html_engine_freeze (e);
 
-	go_to_begin_of_pre_para (e);
+	go_to_begin_of_para (e);
 
 	line_offset = 0;
 	last_space  = 0;
 	do {
+		HTMLObject *flow;
+
 		line_offset = try_break_this_line (e, line_offset, last_space);
+		flow = e->cursor->object->parent;
 		if (html_cursor_forward (e->cursor, e)
 		    && e->cursor->offset == 0 && html_object_get_length (e->cursor->object)
+		    && !html_object_is_container (e->cursor->object)
+		    && html_clueflow_style_equals (HTML_CLUEFLOW (e->cursor->object->parent), HTML_CLUEFLOW (flow))
 		    && html_object_prev_not_slave (e->cursor->object) == NULL) {
 			if (line_offset < LINE_LEN - 1) {
 				gunichar prev;
@@ -608,6 +710,7 @@ html_engine_next_cell (HTMLEngine *e, gboolean create)
 		if (create && HTML_IS_TABLE (e->cursor->object)) {
 			html_cursor_backward (e->cursor, e);
 			html_engine_insert_table_row (e, TRUE);
+			html_cursor_forward (e->cursor, e);
 		}
 		html_engine_show_cursor (e);
 
@@ -638,3 +741,12 @@ html_engine_prev_cell (HTMLEngine *e)
 
 	return FALSE;
 }
+
+void
+html_engine_set_title (HTMLEngine *e, const gchar *title)
+{
+	if (e->title)
+		g_string_free (e->title, TRUE);
+	e->title = g_string_new (title);
+	g_signal_emit_by_name (e, "title_changed");
+}
